function [hmm, llikes] = hmm_em(cfeats, priors, hmm_init, maxiter, update_mu, update_sigmas, update_transitions)
    % Argumenty wejsciowe
    %   cfeats - cell array [1xR] zawierajaca R przykladow uczacych opisujacych dana cyfre
    %   priors - poczatkowe prawdopodobienstwa wejsc do stanow
    %   hmm_init - poczatkowy model HMM
    %   maxiter - maksymalna liczba iteracji algorytmu EM
    %   update_mu [domyslnie true] - aktualizowac wektory srednich?
    %   update_sigmas [domyslnie true]- aktualizowac macierze kowariancji?
    %   update_transitions [domyslnie true] - aktualizowac macierz przejsc?
    %
    % Wartosci zwracane:
    %   hmm - wysetymowany model HMM
    %   llikes - historia log-prawdopodobienstwa w poszczegolnych iteracjachifififi
   
    llike_min_improvement = .5; % minimalna poprawa pomiedzy krokami E i M aby kontynuowac uczenie
    llikes = [-Inf];          % log-prawdpodobienstwa akustyczne w kolejnych iteracjach
    converged = false;
   
    if nargin < 5 update_mu = true; end
    if nargin < 6 update_sigmas = true; end    
    if nargin < 7 update_transitions = true; end 

    [xxx, R] = size(cfeats);
    [F, M, D] = size(hmm_init.mu);
    
    % inicjalizacja akumulatorow zbierajacych statystyki kroku E (uzyte w kroku M do aktualizacji parametrow)
    tot_gammas = zeros(F, 1);
    accs_mu = zeros(F, D); 
    accs_sigma = zeros(F, D, D);
    accs_trans = zeros(F, F);
    
    hmm = hmm_init; iter = 0; 
    while ( (iter < maxiter) && ~converged )
      
        llike = 0.0; 
        RR = R; % tablica cfeats moze zawierac puste komorki, wtedy 
                % RR jest aktualizowane w przypadku opuszczenia przykladu

        %Krok E
        for r=1:R % oblicz statysytki dla kazdej z r wypowiedzi w zbiorze uczacym
        
            o = cell2mat(cfeats(1,r));  
            if is_empty_cell(o) RR=RR-1; continue; end
            T = size(o, 1);
            
            % oblicz prawdopodobienstwa akustyczne z jakim aktualny model hmm 
            % poszczegolne ramki obserwacji wypowiedzi r
            B = compute_pdfs(o, hmm);
            
            % oblicz prawdopodobienstwa w przod i w tyl oraz prawdopodobienstwo gamma
            [llike_r, alphas, betas, gammas] = hmm_fwdback(priors, B, hmm.A, true);
            
            % zakumuluj statystyki do aktualizacji parametrow modelu w kroku M             
            tot_gammas = tot_gammas + sum(gammas, 2); % sum_{t=1}^Tr \gamma_j
            
            if update_mu % dla wektorow srednich
                accs_mu = accs_mu + gammas*o; % sum_{r=1}^R sum_{t=1}^T \gamma_f*o^r
            end  
                     
            if update_sigmas % dla macierzy kowariancji
                for f=1:F
                    mu = reshape(hmm.mu(f,1,:), 1, D); w = zeros(D, D);
                    for t=1:T
                        w = w + gammas(f,t).*(o(t,:)-mu)'*(o(t,:)-mu); %sum_{t=1}^T \gamma_f*(o^r-mu)(o^r-mu)'
                    end
                    accs_sigma(f,:,:) = accs_sigma(f,:,:) + reshape(w, 1, D, D);
                end
            end  
                      
            if update_transitions % dla macierzy przejsc
                 for t=1:T-1
                     accs_trans = accs_trans + hmm.A.*(alphas(:,t)*(B(:,t+1).*betas(:,t+1))');
                 end
            end                  
            llike = llike+llike_r;
        end
        llike = llike./RR;
        fprintf(1, '------ Iteration %d : Log-likelihood is %g.\n', iter, llike);
                 
        %Krok M
        if update_mu
            new_mu = accs_mu./repmat(tot_gammas, 1, D); % rown. 14.52 
            hmm.mu = reshape(new_mu, F, M, D);        
        end
      
        if update_sigmas
            for f=1:F
                new_sigma = reshape(accs_sigma(f,:,:), D , D)./repmat(tot_gammas(f), D, D); %rown. 14.53
                hmm.sigma(f,1,:,:) = diag(diag(new_sigma)); % zakladamy ze estymujemy tylko wariancje 
            end
        end
      
        if update_transitions
            hmm.A = normalise_matrix(accs_trans);
        end
      
        llike_improvement = abs(llike - llikes(end));
        converged = (llike_improvement < llike_min_improvement);
        iter = iter + 1;
        llikes = [llikes llike];
    end    
end
