function [hmm, llikes] = hmm_em_viterbi(cfeats, priors, hmm_init, maxiter)
    % Argumenty wejsciowe
    %   cfeats - cell array [1xR] zawierajaca R przykladow uczacych opisujacych dana cyfre
    %   priors - poczatkowe prawdopodobienstwa wejsc do stanow
    %   hmm_init - poczatkowy model HMM
    %   maxiter - maksymalna liczba iteracji algorytmu EM
    %
    % Wartosci zwracane:
    %   hmm - wysetymowany model HMM
    %   llikes - historia log-prawdopodobienstwa w poszczegolnych iteracjach
   
    llike_min_improvement = .5; % minimalna poprawa pomiedzy krokami E i M aby kontynuowac uczenie
    llikes = [-Inf];          % log-prawdpodobienstwa akustyczne w kolejnych iteracjach
    converged = false;
   
    update_mu = true;          % aktualizowac wektory srednich?
    update_sigmas = true;      % aktualizowac macierze kowariancji?
    update_transitions = true; % aktualizowac macierz przejsc?
   
    [xxx, R]=size(cfeats);
    [F, M, D] =size(hmm_init.mu);
    
    % inicjalizacja akumulatorow zbierajacych statystyki kroku E (uzyte w kroku M do aktualizacji parametrow)
    tot_psis = zeros(F, 1);
    accs_mu = zeros(F, D); 
    accs_sigma = zeros(F, D, D);
    accs_trans = zeros(F, F); %accs_trans(1,1)=1; accs_trans(F,F)=1;
    
    hmm = hmm_init; iter = 0; 
    while ( (iter < maxiter) && ~converged )
        
        llike=0.0;
        
        %Krok E
        for r=1:R % oblicz statysytki dla kazdej z r wypowiedzi w zbiorze uczacym
            o = cell2mat(cfeats(1,r));  
            if is_empty_cell(o) continue; end
            T = size(o, 1);
            % oblicz prawdopodobienstwo akustyczne z jakim aktualny model hmm generuje obserwacje r
            B = compute_pdfs(o, hmm);
            
            % oblicz prawdopodobienstwo najlepszej sciezki
            [llike_r, best_path] = hmm_viterbi(priors, B, hmm.A);
            
            % zbuduj na podstawie best_path macierz psi dla wypowiedzi r (ps. byloby wydajniej zrobic to w hmm_viterbi)
            psi = zeros(F, T);
            for t=1:T
                psi(best_path(t), t) = 1;
                if update_transitions
                    if t>1
                        accs_trans(best_path(t-1), best_path(t)) = accs_trans(best_path(t-1), best_path(t)) + 1;
                    end
                end
            end
            psi
            % zakumuluj statystyki do aktualizacji parametrow modelu   
            tot_psis = tot_psis + sum(psi, 2); 
            
            if update_mu
                accs_mu = accs_mu + psi*o; % sum_{r=1}^R sum_{t=1}^T \psi_f*o^r
            end
            
            if update_sigmas
                for f=1:F
                    mu = reshape(hmm.mu(f,1,:), 1, D); %srednie komponentu f-tego stanu
                    w = zeros(D, D);
                    for t=1:T
                        w = w + psi(f,t).*(o(t,:)-mu)'*(o(t,:)-mu); % (1) sum_{t=1}^T \psi_f*(o^r-mu)(o^r-mu)'
                    end
                    accs_sigma(f,:,:) = accs_sigma(f,:,:) + reshape(w, 1, D, D); %sum_{r=1}^R (1)
                end
            end
                   
            llike = llike+llike_r;
        end
        llike = llike./R;
        fprintf(1, '------ Iteration %d : Log-likelihood is %g.\n', iter, llike);
                 
        %Krok M
        if update_mu
            new_mu = accs_mu./repmat(tot_psis, 1, D); % rown. 14.52 
            hmm.mu = reshape(new_mu, F, M, D);        
        end
      
        if update_sigmas
            for f=1:F
                new_sigma = reshape(accs_sigma(f,:,:), D , D)./repmat(tot_psis(f), D, D); %rown. 14.53
                hmm.sigma(f,1,:,:) = diag(diag(new_sigma)); % zakladamy ze estymujemy tylko wariancje 
            end
      
        if update_transitions
            hmm.A = normalise_matrix(accs_trans./repmat(sum(accs_trans, 2), 1, F));
        end
      
        llike_improvement = abs(llike - llikes(end));
        converged = (llike_improvement < llike_min_improvement);
        iter = iter + 1;
        llikes = [llikes llike];
   end
    
end
