function y = adsl( iter, trType, bitMask, teqIR, Dx, feq, noiseType )
% iter: numer of frame (modem iteration)
% trType: transmision type
%   0 = training signal {1,2,3,4},  no prefix, no TEQ, no FEQ, bitMask ignored
%   1 = SNR estimation: same signal as 0 but: prefix, TEQ, FEQ, bitMask ignored
%   2 = data transmission: all function necessary 
% bitMask: (N/2,1) number of bit transmited in each subchannel 0 means no transmission in this subchannel

Lp = 32;                    % length of prefix
N = 512;                   % frame size sithout prefix
fp = 2.208e6;            % samplig frequence in time domain
sigPower = 23;          % time domain signal energy in dBm
% noiseType = 'AWGN'; % type of noise: AWGN, NBI, Impulse, etc...
noisePower = -100;   % time domain noise power in dBm/Hz
noisePowerNBI = -20;     % NarrowBand noise power in dBm
noiseChannelNBI = 100;
channelNr = 2;         % csaloop: type of communication channel {1...8, 10, 11}
H = zeros( 1, N/2 );
sig = zeros( 1, N/2 );
noise = zeros( 1, N/2 );
ber = zeros( 1, N/2 );
PPAR = [];

% main loop of adsl modem
% one iteration means processing one frame of data
for i = 1:iter
    X1 = genBinData( trType, bitMask );
    X2 = QAM( X1, trType, bitMask );
    [x3, PAR] = Mod( X2, N );
    PPAR = [PPAR, PAR];
    x4 = Prefix( x3, trType, Lp );
    x5 = NormEnergy( x4, sigPower );
    y1 = channel( x5, channelNr );
    y2 = AddNoise( y1, noiseType, noisePower, noisePowerNBI, noiseChannelNBI, fp, iter );
    y3 = DeNormEnergy( y2, sigPower );
    y4 = TEQ( y3, teqIR );
    y5 = DePrefix( y4, trType, N, Dx, Lp );
    Y6 = DeMod( y5, N );
    Y7 = FEQ( Y6, feq );
    Y8 = DeQAM( Y7, trType, bitMask );
    H = HEstim( i, trType, X2, Y6, H );
    [sig, noise] = SNR( i, trType, X2, Y7, sig, noise );
    ber = BER( i, trType, bitMask, X1, Y8, ber );
    %     err = cmp( X1, Y8, X2, Y7 );
end

if( trType == 0 )   % channel estimation
    H( N/2+2:N ) = conj( fliplr( H(2:N/2) ) );
    H = H / (iter-1);
    h = ifft( H );
    h = real(h);
    y = h;
elseif( trType == 1 )   % SNR calculation
    snr = sig./noise;
    y = snr;
elseif( trType == 2 )   % bit transmission
    %error( 'trType = 2: missing code' );
    y = PPAR;
end

%===================== genBinData ====================
function y = genBinData( trType, mask )
% trType: transmission type:
%   0 = training signal {1,2,3,4},  no prefix, no TEQ, no FEQ, bitMask ignored
%   1 = SNR estimation: same signal as 0 but: prefix, TEQ, FEQ, bitMask ignored
%   2 = data transmission
% mask(N/2,1): number of bits in each subchannel
% y(N/2,1): output vector of random data of size mask
persistent state;
if isempty( state )  % avoid correlation between random noise and random data
    rand( 'state', 111 );
    disp( 'initialize persistent random data generator' );
else
    rand( 'state', state );
end

if( trType == 0 || trType == 1 )  % ReverbSequence (ADSL norm C-REVERB1 p.92)
    N = 512;
    b = zeros( N, 1 );
    b( 1:9 ) = 1;
    for i =10:length(b)
        b(i) = xor( b(i-4), b(i-9) );
    end

    for i = 1:N/2
        y( i ) = ( 2 * b( 2*(i-1)+1 ) + b( 2*(i-1)+2 ) ) +1;
    end
elseif( trType == 2 )
    %error( 'genBinData: missing code' );
    for s=1:length(mask)
        if( mask(s) == 0 )
            y(s) = 0;
        else
            y(s) = randint( 1, 1, 2^mask(s) );
        end
    end
else
    y = zeros( size(mask) );
end
state = rand( 'state' );


%===================== QAM ====================
function y = QAM( x, trType, mask )
% x(N/2,1): binary data {0, 1, 2, 3, 4, 5, ... 2^n-1}
% mask(N/2,1): number of bits in each subchannel
% y(N/2,1): QAM
y=zeros( size(mask) );
if( trType == 0 || trType == 1 )    % channel and SNR estimation, only QAM4 used
    QAM4 = [ 1+j, 1-j; -1+j, -1-j ];
    y = zeros( size( mask ) );
    for i = 1:length( mask )
        if( mask(i) ~= 2 )
            y(i) = 0;
        else
            y(i) = QAM4( x(i) );
        end
    end
elseif( trType == 2 )       % QAM4, QAM8, ..., QAM32768
    qamMax = [ 0 1 3 3 5 7 11 15 23 31 47 63 95 127 191 ];
    %error( 'QAM: missing code' );
    for s=1:length(mask)
        M = 2^mask(s);
        if( mask( s ) == 0 )
            y(s) = 0;
        else
            y(s) = qammod( x(s), M ) / qamMax( mask(s) );
        end
    end
else
    y = zeros( size(mask) );
end


%===================== Mod ====================
function [y, PAR] = Mod( x, N )
% x(N/2,1): QAM
% N: length of frame (512 for adsl)
% y(N,1): modulated data (ifft for adsl)
% tip: y have to be real, before modulation assure conjucate symetry
%       output signal have to be normalized to var=1
%error( 'Mod: missing code' );
x(1) = sqrt( 2 );
x( N/2+1 ) = sqrt( 2 );
% x(1) = 0;
% x( N/2+1 ) = 0;

x( N/2+2: N ) = conj( fliplr( x(2:N/2 ) ) );
y = ifft( x );
y = y.*sqrt(N/2);
PAR = max( abs( y ) );

%===================== prefix ====================
function y = Prefix( x, trType, Lp )
% x(N,1): modulated data (in time domain)
% type: transmission type
%     0: training = no prefix
%     1,2: transmission = required prefix
%error( 'Mod: missing code' );
if( trType == 0 )
    y = x;
else
    y = [ x(end-Lp+1:end), x ];
end


%===================== NormEnergy ===============
function y = NormEnergy( x, power )
% x(N+Lp,1): one frame of data in time domain
% power: power of signal in dBm
%error( 'NormEnergy: missing code' );
inputImp = 100;
signalPower = inputImp * 0.001 * 10^( power/10 );
y = x .* sqrt( signalPower );


%===================== channel ==================
function y = channel( x, nr )
% x(N+Lp,1): one frame of data before channel
% nr: simulated type of channel {1...8, 10, 11}
% y(N+Lp,1): one frame of data after channel
persistent ir_residue;
persistent ir;
if isempty( ir_residue )
    load ir;                                    % load impulse response of channel
    eval( sprintf( 'ir=ir%d;', nr ) );
    ir_residue = zeros( length( ir ) -1, 1 );
    disp( 'initialize persistent ir_residue array' );
end
[ y ir_residue ] = filter( ir, 1, x, ir_residue );


%===================== noise ===================
function y = AddNoise( x, type, power, powerNBI, channelNBI, fs, iter )
% x(N+Lp,1)
% type: type of noise: {AWGN, NBI, Impulse}
% power: power of noise in dBm/Hz
% fp: sampling frequency
persistent state;
persistent NBI;
if isempty( state )  % avoid correlation between random noise and random data
    rand( 'state', 112 );
    disp( 'initialize persistent random noise' );
    NBI = narrowNoiseGen( fs, iter*544, channelNBI/256/2*fs, 'AM-FIR',  fs/512, 12345 )';
else
    rand( 'state', state );
end
y = x;
if( strcmp(type,'AWGN') || strcmp( type, 'AWGN+NBI' ) )
    %error( 'AddNoise - AWGN: missing code' );
    inputImp = 100; 
    npower = inputImp * 0.001 * fs/2 * 10^( power/10 );
    xn = randn( size(x) );
    y = y + xn .* sqrt( npower );
end
if( strcmp( type, 'NBI' ) || strcmp ( type, 'AWGN+NBI' ) )
    inputImp = 100; 
    noiseNotchGain = inputImp * 0.001 * 10^( powerNBI/10 );
    y = y + NBI( 1+(iter-1)*length(x) : iter*length(x) ) * sqrt( noiseNotchGain );
end
if( strcmp( type, 'Impulse' ) )
    %error( 'AddNoise - Impulse noise: missing code' );
    y = y;
end
state = rand( 'state' );


%===================== NormEnergy ===============
function y = DeNormEnergy( x, power )
% x(N+Lp,1)
% y(N+Lp,1)
%error( 'DeNormEnergy: missing code' );
inputImp = 100;
signalPower = inputImp * 0.001 * 10^( power/10 );
y = x ./ sqrt( signalPower );


%===================== TEQ ====================
function y = TEQ( x, ir )
% x(N+Lp,1)
% ir(Le,1): impulse response of Time EQualizer
% y(N+Lp,1): equalized frame of data in time domain
persistent eq_residue;
if isempty( eq_residue )
    eq_residue = zeros( length( ir ) -1, 1 );
    disp( 'initialize persistent eq_residue array' );
end
[ y eq_residue ] = filter( ir, 1, x, eq_residue );


%===================== DePrefix =================
function y = DePrefix( x, trType, N, D, Lp )
% x(N+Lp,1)
% type: transmission type
%     0: training = no prefix
%     1: transmission = required prefix
% N: length of data frame in time domain
% D: delay of transmission caused by channel and TEQ
% Lp: prefix length
% y(N,1): one frame of data in time domain without prefix
persistent lastFrame;
if isempty( lastFrame )
    lastFrame = zeros( size( x ) );
    disp( 'initialize last frame delay' );
end

cx = [ lastFrame, x ];
lastFrame = x;

if( trType == 0 )
    y = x;
else
    %error( 'DePrefix: missing code' );
    y = cx( D+Lp+1 : N+D+Lp );
end

%===================== DeMod ==================
function y = DeMod( x, length )
% x(N,1)
% length: lenght of data frame in time domain
% y(N/2,1): complex data frame in frequency domain (demodulated)
%error( 'DeMod: missing code' );
y = fft( x );
y = y( 1:length/2 );
y = y ./ sqrt( length/2 );


%===================== FEQ ====================
function y = FEQ( x, feq )
% x(N/2,1): complex
% feq(N/2,1): Frequency EQualizer
%error( 'FEQ: missing code' );
y = x .* feq;


%===================== DeQAM ==================
function y = DeQAM( x, trType, mask )
% x(N/2,1): equalized data in frequency domain
% mask(N/2,1): number of bits in each subchannel
% y(N/2,1): binary data

if( trType == 0 || trType == 1 )
    %QAM4 = [ 1+j, 1-j; -1+j, -1-j ];
    x1 = round( x );
    for i = 1:length( x )
        rx = real( x1(i) );
        ix = imag( x1(i) );
        if( rx==1 && ix==1 )
            y(i) = 1;
        elseif( rx==1 && ix==-1 )
            y(i) = 2;
        elseif( rx==-1 && ix==1 )
            y(i) = 3;
        elseif( rx==-1 && ix==-1 )
            y(i) = 4;
        else
            y(i) = 0;
        end
    end
elseif( trType == 2 )
    qamMax = [ 0 1 3 3 5 7 11 15 23 31 47 63 95 127 191 ];
    %error( 'DeQAM: missing code' );
    for s=1:length(mask)
        M = 2^mask(s);
        if( mask( s ) == 0 )
            y(s) = 0;
        else
            y(s) = qamdemod( x(s)*qamMax( mask(s) ), M );
        end
    end
else
    y=zeros( size(mask) );
end


%===================== H estimation ===============
function H = HEstim( i, trType, X, Y, Hp )
% estimation of channel impulse response, first frame is always corrupted (skip it!)
% iter: iteration number
% type: transmission type
%     0: training = channel identyfication
%     1: snr estimation = do nothing
%     2: bit transmission = do nothing
% X(N/2,1): transmited data (in frequency domain)
% Y(N/2,1): received data (in frequency domain)
% Hp(N/2,1): previously calculated frequency response of channel
% H(N/2,1): current (accumulated) frequency response of channel
if( trType ~= 0 )
    H = Hp;
    return;
else
    %error( 'HEstim: missing code' );
    if( i == 1 )
        H = zeros( size(Hp) );
    else
        H = Hp + Y./X;
    end
end

function [sig, noise] = SNR( i, trType, X, Y, sigPrev, noisePrev )
% estimation of Signal-to-Noise-Ratio, first frame is always corrupted (skip it!)
% iter: iteration number
% type: transmission type
%     0: training = do nothing
%     1: snr estimation
%     2: bit transmission = do nothing
% X(N/2,1): transmited data (in frequency domain)
% Y(N/2,1): received data (in frequency domain after FEQ)
% sigPrev: previously calculated signal power
% noisePrev: previously calculated noise power
% sig: current (accumulated) signal power
% noise: current (accumulated) noise power
if( trType ~=1 )
        sig = zeros( size( sigPrev ) );
        noise = zeros( size( noisePrev ) );
        return;
else
    %error( 'SNR: missing code' );
    if( i == 1)         % first frame
        sig = zeros( size( sigPrev ) );
        noise = zeros( size( noisePrev ) );
    else
        e = Y - X;
        e = e.*conj(e);

        sig = sigPrev + (X.*conj(X));
        noise = noisePrev + e;
    end
end


function ber = BER( i, trType, mask, X, Y, berp )
% estimation of BitErrorRatio, first frame is always corrupted (skip it!)
% type: transmission type
%     0: training = do nothing
%     1: snr estimation = do nothing
%     2: bit transmission
% mask: number of bits in subchannels
% X(N/2,1): transmited data (binary numbers)
% Y(N/2,1): received data (binary numbers)
persistent lastX;
% persistent xx;
% persistent yy;
if isempty( lastX )
    lastX = zeros( size( X ) );
    disp( 'initialize last frame delay' );
end

if( trType ~=2 )
    ber = zeros( size(X) );
    return;
else
    %error( 'BER: missing code' );
    if( i<2 )
        ber = zeros( size(X) );
    else
        ber = berp + ( (lastX-Y) ~= 0 );
    end    
end
lastX = X;

%===================== cmp ====================
function err = cmp( x1, y1, x2, y2 )
% compare and plot, etc..
persistent lastX2;
% persistent xx;
% persistent yy;
if isempty( lastX2 )
    lastX2 = zeros( size( x2 ) );
    disp( 'initialize last frame delay' );
    err =  zeros( size(x2) );
    return;
end

figure(1);
t = 1:length(y2);
plot( t, real( lastX2(t)), 'b.', t, real(y2(t)), 'bo', t, imag( lastX2(t)), 'rx', t, imag(y2(t)), 'ro' );
err = zeros( size(x1) );
lastX2 = x2;


